//=============================================================================
// INSTRUCTION CLASS
//=============================================================================

#include "instructions.h"
#include "opcodes.h"

#include "assemble.h"
#include "streams.h"
#include "labels.h"


//-----------------------------------------------------------------------------
DWORD INSTRUCTION::get_reg(PARAMLIST *pl)
{
	if(assemble.config.paramtype_check)
	{
		if(pl->type != PTYPE_REG) return(ERR_WARNING|RETC_WRONG_PTYPE);
	}
	else
	{
		if((pl->type != PTYPE_REG)&&(pl->type != PTYPE_VALUE)) return(ERR_WARNING|RETC_WRONG_PTYPE);
	}
	if((pl->val < 0)||(pl->val > 31)) return(ERR_WARNING|RETC_REG_OVRFLW);
	return(RETC_OK);
}
//-----------------------------------------------------------------------------
DWORD INSTRUCTION::get_v32(PARAMLIST *pl)
{
	if(/*(pl->type != PTYPE_REG)&&*/(pl->type != PTYPE_VALUE)) return(ERR_WARNING|RETC_WRONG_PTYPE);
	if((pl->val < 0)||(pl->val > 31)) return(ERR_WARNING|RETC_REG_OVRFLW);
	return(RETC_OK);
}
//-----------------------------------------------------------------------------
DWORD INSTRUCTION::get_imm(PARAMLIST *pl)
{
	if((pl->type != PTYPE_VALUE)&&(pl->type != PTYPE_LBL)) return(ERR_WARNING|RETC_WRONG_PTYPE);
	if((pl->val < -0xFFFF)||(pl->val > 0xFFFF)) return(ERR_WARNING|RETC_IMM_OVRFLW);
	pl->val = ((__int16)pl->val)&0xffff;
	return(RETC_OK);
}
//-----------------------------------------------------------------------------
DWORD INSTRUCTION::get_inv(PARAMLIST *pl)
{
	if((pl->type != PTYPE_VALUE)&&(pl->type != PTYPE_LBL)) return(ERR_WARNING|RETC_WRONG_PTYPE);
	if((pl->val < -0xFFFF)||(pl->val > 0xFFFF)) return(ERR_WARNING|RETC_IMM_OVRFLW);
	pl->val = (-(__int16)pl->val)&0xffff;
	return(RETC_OK);
}
//-----------------------------------------------------------------------------
DWORD INSTRUCTION::get_off(PARAMLIST *pl)
{
	if((pl->type != PTYPE_VALUE)&&(pl->type != PTYPE_LBL)) return(ERR_WARNING|RETC_WRONG_PTYPE);
	if(pl->val%4) return(RETC_OFF_INVALID);
	pl->val = (pl->val - (assemble.prog_cntr+4))>>2;
	if((pl->val < -0x8000)||(pl->val >= 0x8000)) return(ERR_WARNING|RETC_OFF_OVRFLW);
	pl->val = ((__int16)pl->val)&0xffff;
	return(RETC_OK);
}
//-----------------------------------------------------------------------------
DWORD INSTRUCTION::get_nobase(PARAMLIST *pl)
{
	if((pl->type != PTYPE_VALUE)&&(pl->type != PTYPE_LBL)) return(ERR_WARNING|RETC_WRONG_PTYPE);
	//pl->val = (pl->val - (assemble.prog_cntr/*+4*/));	//def base=gp -> make ofs pc_rel
	if((pl->val < -0x8000)||(pl->val >= 0x8000)) return(ERR_WARNING|RETC_OFF_OVRFLW);
	pl->val = ((__int16)pl->val)&0xffff;
	return(RETC_OK);
}
//-----------------------------------------------------------------------------
DWORD INSTRUCTION::get_off_base(PARAMLIST *off, PARAMLIST *base)
{
	if(((off->type != PTYPE_OFF)&&(off->type != PTYPE_DEFOFF))
		||(base->type != PTYPE_BASE)) return(ERR_WARNING|RETC_WRONG_PTYPE);
	if((base->val < 0)||(base->val > 31)) return(ERR_WARNING|RETC_REG_OVRFLW);
	if(off->type == PTYPE_DEFOFF) off->val=0;	//on default ofs...
	else									//otherwise...
	{
		//if(base->val == 28)
		//	  off->val = (off->val - (assemble.prog_cntr/*+4*/));	//base=gp -> make ofs pc_rel
	}
	if((off->val < -0xFFFF)||(off->val >= 0xFFFF)) return(ERR_WARNING|RETC_OFF_OVRFLW);
	off->val = ((__int16)off->val)&0xffff;
	return(RETC_OK);
}
//-----------------------------------------------------------------------------
// INSTRUCTION::process_instr: write the instr. to file
//-----------------------------------------------------------------------------
DWORD INSTRUCTION::process_instr(int instr, LIST *plst)
{
	dbgout("INSTRUCTION::process_instr(INT[" << instr << "], LIST[" << (DWORD)plst << "])\n")

	PARAMLIST *pl[3];

	pl[0] = (PARAMLIST*)plst->get(0);
	pl[1] = (PARAMLIST*)plst->get(1);
	pl[2] = (PARAMLIST*)plst->get(2);

	DWORD ret;

	#define _ERROR(num) do_error(num, ERR_SHOW_POS);

	DWORD opcode = opcodes[instr].opcode;
	switch(opcodes[instr].param)
	{
	case PARAM_RD_RS_RT:
		if((ret=get_reg(pl[0])) != RETC_OK) _ERROR(ret+1);
		if((ret=get_reg(pl[1])) != RETC_OK) _ERROR(ret+2);
		if((ret=get_reg(pl[2])) != RETC_OK) _ERROR(ret+3);
		opcode=opcode|_PARAM(0,_RD)|_PARAM(1,_RS)|_PARAM(2,_RT);
		break;
	case PARAM_RDRS_RT: //rd and rs are pointing to the same register
		if((ret=get_reg(pl[0])) != RETC_OK) _ERROR(ret+1);
		if((ret=get_reg(pl[1])) != RETC_OK) _ERROR(ret+2);
		opcode=opcode|_PARAM(0,_RD)|_PARAM(0,_RS)|_PARAM(1,_RT);
		break;

	case PARAM_RT_RS_IMM:
		if((ret=get_reg(pl[0])) != RETC_OK) _ERROR(ret+1);
		if((ret=get_reg(pl[1])) != RETC_OK) _ERROR(ret+2);
		if((ret=get_imm(pl[2])) != RETC_OK) _ERROR(ret+3);
		opcode=opcode|_PARAM(0,_RT)|_PARAM(1,_RS)|_PARAM(2,_IMM);
		break;
	case PARAM_RTRS_IMM: //rt and rs are pointing to the same register
		if((ret=get_reg(pl[0])) != RETC_OK) _ERROR(ret+1);
		if((ret=get_imm(pl[1])) != RETC_OK) _ERROR(ret+2);
		opcode=opcode|_PARAM(0,_RT)|_PARAM(0,_RS)|_PARAM(1,_IMM);
		break;

	case PARAM_RT_RS_INV:
		if((ret=get_reg(pl[0])) != RETC_OK) _ERROR(ret+1);
		if((ret=get_reg(pl[1])) != RETC_OK) _ERROR(ret+2);
		if((ret=get_inv(pl[2])) != RETC_OK) _ERROR(ret+3);
		opcode=opcode|_PARAM(0,_RT)|_PARAM(1,_RS)|_PARAM(2,_INV);
		break;
	case PARAM_RTRS_INV: //rt and rs are pointing to the same register
		if((ret=get_reg(pl[0])) != RETC_OK) _ERROR(ret+1);
		if((ret=get_inv(pl[1])) != RETC_OK) _ERROR(ret+2);
		opcode=opcode|_PARAM(0,_RT)|_PARAM(0,_RS)|_PARAM(1,_INV);
		break;

	case PARAM_RS_RT_OFF:
		if((ret=get_reg(pl[0])) != RETC_OK) _ERROR(ret+1);
		if((ret=get_reg(pl[1])) != RETC_OK) _ERROR(ret+2);
		if((ret=get_off(pl[2])) != RETC_OK) _ERROR(ret+3);
		opcode=opcode|_PARAM(0,_RS)|_PARAM(1,_RT)|_PARAM(2,_OFF);
		break;

	case PARAM_RS_OFF:
		if((ret=get_reg(pl[0])) != RETC_OK) _ERROR(ret+1);
		if((ret=get_off(pl[1])) != RETC_OK) _ERROR(ret+2);
		opcode=opcode|_PARAM(0,_RS)|_PARAM(1,_OFF);
		break;

	case PARAM_NOTHING:
		//do nothing
		break;

	case PARAM_CODE:
		if((pl[0]->type != PTYPE_VALUE)&&(pl[0]->type != PTYPE_LBL)) _ERROR(ERR_WARNING|RETC_WRONG_PTYPE);
		if((pl[0]->val < 0)||(pl[0]->val > 0xFFFFF)) _ERROR(ERR_WARNING|RETC_CODE_OVRFLW);
		opcode=opcode|_PARAM(0,_CODE);
		break;

	case PARAM_RS_RT:
		if((ret=get_reg(pl[0])) != RETC_OK) _ERROR(ret+1);
		if((ret=get_reg(pl[1])) != RETC_OK) _ERROR(ret+2);
		opcode=opcode|_PARAM(0,_RS)|_PARAM(1,_RT);
		break;

	case PARAM_RD_RT_SA:
		if((ret=get_reg(pl[0])) != RETC_OK) _ERROR(ret+1);
		if((ret=get_reg(pl[1])) != RETC_OK) _ERROR(ret+2);
		if((ret=get_v32(pl[2])) != RETC_OK) _ERROR(ret+3);
		opcode=opcode|_PARAM(0,_RD)|_PARAM(1,_RT)|_PARAM(2,_SA);
		break;
	case PARAM_RDRT_SA: //rd and rt are pointing to the same register
		if((ret=get_reg(pl[0])) != RETC_OK) _ERROR(ret+1);
		if((ret=get_v32(pl[1])) != RETC_OK) _ERROR(ret+2);
		opcode=opcode|_PARAM(0,_RD)|_PARAM(0,_RT)|_PARAM(1,_SA);
		break;

	case PARAM_RD_RT_RS:
		if((ret=get_reg(pl[0])) != RETC_OK) _ERROR(ret+1);
		if((ret=get_reg(pl[1])) != RETC_OK) _ERROR(ret+2);
		if((ret=get_reg(pl[2])) != RETC_OK) _ERROR(ret+3);
		opcode=opcode|_PARAM(0,_RD)|_PARAM(1,_RT)|_PARAM(2,_RS);
		break;
	case PARAM_RDRT_RS: //rd and rt are pointing to the same register
		if((ret=get_reg(pl[0])) != RETC_OK) _ERROR(ret+1);
		if((ret=get_reg(pl[1])) != RETC_OK) _ERROR(ret+2);
		opcode=opcode|_PARAM(0,_RD)|_PARAM(0,_RT)|_PARAM(1,_RS);
		break;

	case PARAM_TARG:
		if((pl[0]->type != PTYPE_VALUE)&&(pl[0]->type != PTYPE_LBL)) _ERROR(ERR_WARNING|RETC_WRONG_PTYPE);
		if(pl[0]->val%4) _ERROR(ERR_WARNING|RETC_TARG_INVALID);
		if((pl[0]->val & (!0xFFFFFFF)) != ((assemble.prog_cntr+4) & (!0xFFFFFFF))) _ERROR(RETC_NC_REGION);
		pl[0]->val = (pl[0]->val & 0xFFFFFFF) >> 2;
		if((pl[0]->val < 0)||(pl[0]->val > 0x3ffffff)) _ERROR(ERR_WARNING|RETC_TARG_OVRFLW);
		opcode=opcode|_PARAM(0,_TARG);
		break;

	case PARAM_RS:
		if((ret=get_reg(pl[0])) != RETC_OK) _ERROR(ret);
		opcode=opcode|_PARAM(0,_RS);
		break;

	case PARAM_RD_RS:
		if((ret=get_reg(pl[0])) != RETC_OK) _ERROR(ret+1);
		if((ret=get_reg(pl[1])) != RETC_OK) _ERROR(ret+2);
		opcode=opcode|_PARAM(0,_RD)|_PARAM(1,_RS);
		break;

  case PARAM_OP_OFF_BASE_:	//OP=IMM5=FT=RT
	case PARAM_IMM5_OFF_BASE_:
  case PARAM_FT_OFF_BASE_:
	case PARAM_RT_OFF_BASE_:
		if((ret=get_reg(pl[0])) != RETC_OK) _ERROR(ret+1);
		if((ret=get_off_base(pl[1], pl[2])) != RETC_OK) _ERROR(ret+2);
		opcode=opcode|_PARAM(0,_RT)|_PARAM(1,_OFF)|_PARAM(2,_BASE);
		break;

  case PARAM_OP_NOBASE:	//OP=IMM5=FT=RT
	case PARAM_IMM5_NOBASE:	//produces the code 4 an off(base)-instr. using default 4 base
  case PARAM_FT_NOBASE:
	case PARAM_RT_NOBASE:
		if((ret=get_reg(pl[0])) != RETC_OK) _ERROR(ret+1);
		if((ret=get_nobase(pl[1])) != RETC_OK) _ERROR(ret+2);
		opcode=opcode|_PARAM(0,_RT)|_PARAM(1,_NOBASE);
		break;

	case PARAM_RT_IMM:
		if((ret=get_reg(pl[0])) != RETC_OK) _ERROR(ret+1);
		if((ret=get_imm(pl[1])) != RETC_OK) _ERROR(ret+2);
		opcode=opcode|_PARAM(0,_RT)|_PARAM(1,_IMM);
		break;

	case PARAM_RD:
		if((ret=get_reg(pl[0])) != RETC_OK) _ERROR(ret);
		opcode=opcode|_PARAM(0,_RD);
		break;

	case PARAM_RS_IMM:
		if((ret=get_reg(pl[0])) != RETC_OK) _ERROR(ret+1);
		if((ret=get_imm(pl[1])) != RETC_OK) _ERROR(ret+2);
		opcode=opcode|_PARAM(0,_RS)|_PARAM(1,_IMM);
		break;

	case PARAM_RD_RT:
		if((ret=get_reg(pl[0])) != RETC_OK) _ERROR(ret+1);
		if((ret=get_reg(pl[1])) != RETC_OK) _ERROR(ret+2);
		opcode=opcode|_PARAM(0,_RD)|_PARAM(1,_RT);
		break;

    case PARAM_OFF:
		if((ret=get_off(pl[0])) != RETC_OK) _ERROR(ret);
		opcode=opcode|_PARAM(0,_OFF);
		break;

    case PARAM_RT:
		if((ret=get_reg(pl[0])) != RETC_OK) _ERROR(ret);
		opcode=opcode|_PARAM(0,_RT);
		break;

    case PARAM_RT_RD:
		if((ret=get_reg(pl[0])) != RETC_OK) _ERROR(ret+1);
		if((ret=get_reg(pl[1])) != RETC_OK) _ERROR(ret+2);
		opcode=opcode|_PARAM(0,_RT)|_PARAM(1,_RD);
		break;

    case PARAM_RT_REG:
		if((ret=get_reg(pl[0])) != RETC_OK) _ERROR(ret+1);
		if((ret=get_reg(pl[1])) != RETC_OK) _ERROR(ret+2);
		opcode=opcode|_PARAM(0,_RT)|_PARAM(1,_REG);
		break;

    case PARAM_FD_FS:
		if((ret=get_reg(pl[0])) != RETC_OK) _ERROR(ret+1);
		if((ret=get_reg(pl[1])) != RETC_OK) _ERROR(ret+2);
		opcode=opcode|_PARAM(0,_FD)|_PARAM(1,_FS);
		break;

    case PARAM_FD_FS_FT:
		if((ret=get_reg(pl[0])) != RETC_OK) _ERROR(ret+1);
		if((ret=get_reg(pl[1])) != RETC_OK) _ERROR(ret+2);
		if((ret=get_reg(pl[2])) != RETC_OK) _ERROR(ret+3);
		opcode=opcode|_PARAM(0,_FD)|_PARAM(1,_FS)|_PARAM(2,_FT);
		break;

    case PARAM_FS_FT:
		if((ret=get_reg(pl[0])) != RETC_OK) _ERROR(ret+1);
		if((ret=get_reg(pl[1])) != RETC_OK) _ERROR(ret+2);
		opcode=opcode|_PARAM(0,_FS)|_PARAM(1,_FT);
		break;

    case PARAM_RT_FS:
		if((ret=get_reg(pl[0])) != RETC_OK) _ERROR(ret+1);
		if((ret=get_reg(pl[1])) != RETC_OK) _ERROR(ret+2);
		opcode=opcode|_PARAM(0,_RT)|_PARAM(1,_FS);
		break;

    case PARAM_FD_FT:
		if((ret=get_reg(pl[0])) != RETC_OK) _ERROR(ret+1);
		if((ret=get_reg(pl[1])) != RETC_OK) _ERROR(ret+2);
		opcode=opcode|_PARAM(0,_FD)|_PARAM(1,_FT);
		break;

    case PARAM_RT_ID:
		if((ret=get_reg(pl[0])) != RETC_OK) _ERROR(ret+1);
		if((ret=get_reg(pl[1])) != RETC_OK) _ERROR(ret+2);
		opcode=opcode|_PARAM(0,_RT)|_PARAM(1,_ID);
		break;

    case PARAM_RT_FD:
		if((ret=get_reg(pl[0])) != RETC_OK) _ERROR(ret+1);
		if((ret=get_reg(pl[1])) != RETC_OK) _ERROR(ret+2);
		opcode=opcode|_PARAM(0,_RT)|_PARAM(1,_FD);
		break;

	default: return(RETC_UNKWN_INSTR);
	}

	//write listing
	if((!assemble.preprocess)&&(assemble.lst_fh))
	{
#ifdef __GNUC__
        fprintf(assemble.lst_fh, "%08llX\t%08X\t\t",
			assemble.prog_cntr, (DWORD)opcode);
#else
		fprintf(assemble.lst_fh, "%08I64X\t%08X\t\t",
			assemble.prog_cntr, (DWORD)opcode);
#endif
		if(opcodes[instr].pnum>0)
			fprintf(assemble.lst_fh, "%-7s ", opcodes[instr].name);
		else
			fprintf(assemble.lst_fh, "%s", opcodes[instr].name);
        int cnt;
		for(cnt=0; cnt<opcodes[instr].pnum-1; cnt++)
		{
#ifdef __GNUC__
			if(pl[cnt]->str) fprintf(assemble.lst_fh, "%02llX [%s], ", pl[cnt]->val, pl[cnt]->str);
			else fprintf(assemble.lst_fh, "%02llX, ", pl[cnt]->val);
#else
			if(pl[cnt]->str) fprintf(assemble.lst_fh, "%02I64X [%s], ", pl[cnt]->val, pl[cnt]->str);
			else fprintf(assemble.lst_fh, "%02I64X, ", pl[cnt]->val);
#endif
		}

		if(opcodes[instr].pnum>0)
		{
#ifdef __GNUC__
			if(pl[cnt]->str) fprintf(assemble.lst_fh, "%02llX [%s]", pl[cnt]->val, pl[cnt]->str);
			else fprintf(assemble.lst_fh, "%02llX", pl[cnt]->val);
#else
			if(pl[cnt]->str) fprintf(assemble.lst_fh, "%02I64X [%s]", pl[cnt]->val, pl[cnt]->str);
			else fprintf(assemble.lst_fh, "%02I64X", pl[cnt]->val);
#endif
		}

		fprintf(assemble.lst_fh, "\n");
	}

	//write to file and return
	stream.out.opcode(opcode, 4);
	return(RETC_OK);
}


//-----------------------------------------------------------------------------
// INSTRUCTION::exec: scans str for valid instr
//-----------------------------------------------------------------------------
DWORD INSTRUCTION::exec(char *instr, char *args)
{
	PARAM param( PARAM_GET_VALUES|PARAM_GET_STRINGS );
	PARAMLIST *pl;

	DWORD ret, pnum = 0;	//return value and number of params
	BOOL is_targ = false;	//compare targets
	BOOL found1 = false;	//found at least one matching instruction str
	int instr_cnt = 0;	//number of instruction in tbl

	dbgout("INSTRUCTION::exec(CHAR[" << instr << "], CHAR[" << args << "])\n")

    if(assemble.config.iset == SEL_ISET_VPU1)
    {
        //in this mode we ignore opcodes with more than one parameter types
        //if we found the first matching opcode name we would use this without comparing
        //number of parameters cause we have to search loweer instruction opcode with this param-type

        for(;instr_cnt < NUM_OPCODES; instr_cnt++)
        {
            if(assemble.cmp(instr, opcodes[instr_cnt].name))
            {
                is_targ = true;
                //could i realy do this...? only if lower instructions r all different to upper
                //if(opcodes[instr_cnt].target == TARGET_VU1_L) return(RETC_UPPER_VPU_CODE_MISSING);

                if(opcodes[instr_cnt].target == TARGET_VU1_U)
                {
                    found1 = true;
                }
            }
        }
    } else {
        for(;instr_cnt<NUM_OPCODES;instr_cnt++) if(assemble.cmp(instr, opcodes[instr_cnt].name))
        {
            is_targ=true;
            if(( (assemble.config.iset == SEL_ISET_R3000) && (opcodes[instr_cnt].target == TARGET_R3000_R5900)) ||
                ( (assemble.config.iset == SEL_ISET_R3000) && (opcodes[instr_cnt].target == TARGET_R3000      )) ||
                ( (assemble.config.iset == SEL_ISET_R5900) && (opcodes[instr_cnt].target == TARGET_R3000_R5900)) ||
                ( (assemble.config.iset == SEL_ISET_R5900) && (opcodes[instr_cnt].target == TARGET_R5900      )) ||
                ( (assemble.config.iset == SEL_ISET_VPU0 ) && (opcodes[instr_cnt].target == TARGET_VU0        )) )
            {
                found1=true;

                if(assemble.preprocess)
                {
                    stream.out.align(4);		//make code adressable
                    label.preset(assemble.prog_cntr, LBL_TYPE_CODE);	//update label 4 this line
                    //we could test params right here now
                    stream.out.opcode(NULL, 4); //but we simply think it will be a valid instr
                    return(RETC_OK);
                } else {
                    if(!param.scan_cntr)
                    {
                        stream.out.align(4);		//make code adressable
                        label.preset(assemble.prog_cntr, LBL_TYPE_CODE);	//update label 4 this line
                        if((ret = param.scan(args)) != RETC_OK) return(ret);
                        pnum = param.ParamList->members;
                        while(pnum)
                        {
                            pl = (PARAMLIST*)param.ParamList->get(--pnum);
                            if(!pl) do_error(ERR_FATAL|RETC_LISTREQUEST, 0);
                            if(pl->retc != RETC_OK) return(pl->retc);
                        }
                        pnum = param.ParamList->members;
                    }
                    if(opcodes[instr_cnt].pnum == (int)pnum)
                    {
                        return( process_instr(instr_cnt, param.ParamList) );
                    }
                }
            }
        }
    }

    if(found1)
    {
        return(RETC_UNKWN_INSTR_SYNTAX);
    }
    if(is_targ)
    {
        return(RETC_TARG_MISMATCH);
    }
    return(RETC_UNKWN_INSTR);
}

//-----------------------------------------------------------------------------
// INSTRUCTION::if_exist: test if string is an instruction
//-----------------------------------------------------------------------------
BOOL INSTRUCTION::if_exist(char *instr)
{
    int instr_cnt=0;

    for(;instr_cnt<NUM_OPCODES;instr_cnt++)
        if(assemble.cmp(instr, opcodes[instr_cnt].name)) return(true);

    return(false);
}
