//=============================================================================
// 32=bit assembler for ps2 risc5900mips by wiRe (01/21/2001)
//=============================================================================

#include <iostream>
#include <fstream>
#include "main.h"

//-----------------------------------------------------------------------------
// classes
//-----------------------------------------------------------------------------
#ifdef _DEBUG
fstream fdbg;
#endif

ASSEMBLE assemble;
INSTRUCTION instruction;
DIRECTIVE directive;
INTMATH intmath;
MACRO macro;
STREAM stream;
ASMCLASS asmclass;
LABEL label;

const char _DEFAULT_ERRLOG[] = "errors.log";
const char _PSX_[] = "psx";
const char _SCE_[] = "sce";
const char _ECO_[] = "eco";
const char _CPE_[] = "cpe";
const char _BIN_[] = "bin";
const char _ELF_[] = "elf";

//-----------------------------------------------------------------------------
// main
//-----------------------------------------------------------------------------
int main(int argc, char *argv[], char *envp[])
{
    if (argc >= 2 && strcmp(argv[1], "--nologo") == 0)
      --argc, ++argv;
    else
      cout
//        << "asm5900 v1.8b: ps2 risc5900 mips assembler by wiRe (20/01/2001)"
        << "asm5900 v1.8b: ps2 risc5900 mips assembler by wiRe (2006-02-12-mfx6)"
        << endl
        << "Free software distributed under the terms of the GNU GPL"
        << endl
        << "http://napalm-x.com/~wiRe/?projects.asm5900"
        << endl;

#ifdef _DEBUG
    cout << "DEBUG-VERSION! (write output to dbglog.txt)" << endl;
    fdbg.setbuf(NULL, 0);
    fdbg.open( "DBGLOG.TXT", ios::trunc );
#endif

    //scan parameters
    char *fin = NULL;
    char *fout = NULL;
    char *flog = NULL;
//    BOOL targ = false;
    IM_RET imr;

    if(argc<=1) do_error(ERR_FATAL|RETC_UNKWN_CMDPARAM, ERR_SHOW_NOTHING);
    for(int cntr=1; cntr<argc; cntr++)
    {
        char *p = argv[cntr];
        if (p[0] == '-' && p[1])
        {
	    char *arg = (p[1] && p[2] == ':') ? p + 3 : p + 2;
            switch(p[1])
            {
                case '?':
                case 'h': case 'H':
                    cout << ( "syntax: asm5900 source [/o{:<file>}] [/e{:<file>}] [options]\n"
                        "\n"
                        "/d{:<file>}\t\twrite debug info to file (def: *.dbg)\n"
                        "/e{:<file>}\t\tlog screen output to file (def: errors.log)\n"
                        "/l{:<file>}\t\twrite listing to file (def: *.lst)\n"
                        "/o{:<file>}\t\twrite assembled code to file (def: *.<target>)\n"
                        "/t:<target>\t\tselect target (PSX,SCE,ECO,CPE,BIN,def: BIN)\n"
                        "\n"
                        "/$<-|+>\t\t\tuse $ to define hex numbers (def: $-)\n"
                        "/c<+|->\t\t\ttoggle case sensitive mode (def: c+)\n"
                        "/g<-|+>\t\t\ttoggle global label adressing mode (def: g-)\n"
                        "/r<+|->\t\t\tload default register values (def: r+)\n"
                        "\n"
                        "/i:<path>\t\tsets search path for include files\n"
                        "/m:<value>\t\tsets maximum number of passes (2..99, def: 4)\n"
                        "/s:<symbol>{=<value>}\tdefines a symbol (true on defaul)\n"
                        "/w<-|+>{<warnclass>}\tcontrols the generation of warnings\n"
                        "\n"
                        "/h, /?\t\t\tdisplay this help screen\n"
                        ""
                        );
                    return(0);

                case 'd': case 'D':
                    if(!assemble.config.dbg_file)
                    {
                        if (*arg)
                        {
                            assemble.config.dbg_file = arg;
                            break;
                        }
                        else // otherwise use default
                        {
                            assemble.config.dbg_file = (char *)-1;
                            break;
                        }
                    }
                    do_error(ERR_FATAL|RETC_UNKWN_CMDPARAM, ERR_SHOW_NOTHING);

                case 'e': case 'E':
                    if(!flog)
                    {
                        if (*arg)
                        {
                            flog = arg;
                            break;
                        }
                        else // otherwise use default
                        {
                            flog=(char *)malloc(sizeof(_DEFAULT_ERRLOG)+1);
                            if(!flog) do_error(ERR_FATAL|RETC_OUTOFMEM, 0);
                            strcpy(flog, _DEFAULT_ERRLOG);
                            break;
                        }
                    }
                    do_error(ERR_FATAL|RETC_UNKWN_CMDPARAM, ERR_SHOW_NOTHING);

                case 'i': case 'I':
                    if(*arg)
                    {
                        assemble.config.include_path[assemble.config.include_path_n++] = arg;
                        break;
                    }
                    do_error(ERR_FATAL|RETC_UNKWN_CMDPARAM, ERR_SHOW_NOTHING);

                case 'l': case 'L':
                    if(!assemble.config.lst_file)
                    {
                        if(*arg)
                        {
                            assemble.config.lst_file = arg;
                            break;
                        }
                        else // otherwise use default
                        {
                            assemble.config.lst_file = (char *)-1;
                            break;
                        }
                    }
                    do_error(ERR_FATAL|RETC_UNKWN_CMDPARAM, ERR_SHOW_NOTHING);

                case 'o': case 'O':
                    if(!fout)
                    {
                        if(*arg)
                        {
                            fout = arg;
                            break;
                        }
                        else // otherwise use default
                        {
                            fout = (char *)-1;
                            break;
                        }
                    }
                    do_error(ERR_FATAL|RETC_UNKWN_CMDPARAM, ERR_SHOW_NOTHING);

                case 't': case 'T':
                    if(*arg)
                    {
                        if(assemble.config.target != SEL_TARG_DEFAULT)
                            do_error(ERR_FATAL|RETC_TARG_ALRD_SEL, ERR_SHOW_NOTHING);

                        if(stricmp(arg, _PSX_) == 0) {assemble.config.target = SEL_TARG_PSX; assemble.binary = false; break;}
                        if(stricmp(arg, _SCE_) == 0) {assemble.config.target = SEL_TARG_SCE; assemble.binary = false; break;}
                        if(stricmp(arg, _ECO_) == 0) {assemble.config.target = SEL_TARG_ECO; assemble.binary = false; break;}
                        if(stricmp(arg, _CPE_) == 0) {assemble.config.target = SEL_TARG_CPE; assemble.binary = false; break;}
                        if(stricmp(arg, _BIN_) == 0) {assemble.config.target = SEL_TARG_BIN; assemble.binary = true; break;}
                        if(stricmp(arg, _ELF_) == 0) {assemble.config.target = SEL_TARG_ELF; assemble.binary = false; break;}
                    }
                    do_error(ERR_FATAL|RETC_UNKWN_CMDPARAM, ERR_SHOW_NOTHING);

                case '$':
                    switch(*arg)
                    {
                        case '+': intmath.enable_$ = true; break;
                        case '-': intmath.enable_$ = false; break;
                        default: do_error(ERR_FATAL|RETC_UNKWN_CMDPARAM, ERR_SHOW_NOTHING);
                    }
                    break;

                case 'c': case 'C':
                    switch(*arg)
                    {
                        case '+': assemble.config.case_sensitive = true; break;
                        case '-': assemble.config.case_sensitive = false; break;
                        default: do_error(ERR_FATAL|RETC_UNKWN_CMDPARAM, ERR_SHOW_NOTHING);
                    }
                    break;

                case 'g': case 'G':
                    switch(*arg)
                    {
                        case '+': assemble.config.default_local = false; break;
                        case '-': assemble.config.default_local = true; break;
                        default: do_error(ERR_FATAL|RETC_UNKWN_CMDPARAM, ERR_SHOW_NOTHING);
                    }
                    break;

                case 'r': case 'R':
                    switch(*arg)
                    {
                        case '+': assemble.config.default_define = true; break;
                        case '-': assemble.config.default_define = false; break;
                        default: do_error(ERR_FATAL|RETC_UNKWN_CMDPARAM, ERR_SHOW_NOTHING);
                    }
                    break;

                case 'm': case 'M':
                    imr.str = arg;
                    if((intmath.calc(&imr))&&(imr.value>=2)&&(imr.value<=99))
                    {
                        assemble.config.max_pass = (int)imr.value;
                        break;
                    }
                    do_error(ERR_FATAL|RETC_UNKWN_CMDPARAM, ERR_SHOW_NOTHING);

                case 'q':
                    assemble.config.verbose -= 1;
                    break;
                case 'v':
                    assemble.config.verbose += 1;
                    break;

                default:
                    do_error(ERR_FATAL|RETC_UNKWN_CMDPARAM, ERR_SHOW_NOTHING);

            }
        } else {
            if(fin) do_error(ERR_FATAL|RETC_TOO_MUCH_PARAMS, ERR_SHOW_NOTHING);
            fin = p;
        }
    }


    char *dot;

    if(!fin) do_error(ERR_FATAL|RETC_UNKWN_CMDPARAM, ERR_SHOW_NOTHING);

    if(fout == (char *)-1)	//define default outputfile
    {
        fout=(char *)malloc(strlen(fin)+5);
        if(!fout) do_error(ERR_FATAL|RETC_OUTOFMEM, 0);
        strcpy(fout, fin);
        dot = strrchr(fout, '.');
        if(dot)
        {
            *dot = NULL;
        }
    }

    if(assemble.config.dbg_file == (char *)-1)	//then calc default name
    {
        assemble.config.dbg_file = (char *)malloc(strlen(fin)+5);
        if(!assemble.config.dbg_file) do_error(ERR_FATAL|RETC_OUTOFMEM, 0);
        strcpy(assemble.config.dbg_file, fin);
        dot = strrchr(assemble.config.dbg_file, '.');
        if(!dot) strcat(assemble.config.dbg_file, ".dbg");
        else strcpy(dot, ".dbg");
    }
    if(assemble.config.dbg_file)
    {
        if (assemble.config.verbose >= 2)
            cout << "write debug info to file: " << assemble.config.dbg_file
                 << endl;
    }

    if(assemble.config.lst_file == (char *)-1)      //then calc default name
    {
        assemble.config.lst_file = (char *)malloc(strlen(fin)+5);
        if(!assemble.config.lst_file)
        {
            do_error(ERR_FATAL|RETC_OUTOFMEM, 0);
        }
        strcpy(assemble.config.lst_file, fin);
        dot = strrchr(assemble.config.lst_file, '.');

        if(!dot)
        {
            strcat(assemble.config.lst_file, ".lst");
        } else {
            strcpy(dot, ".lst");
        }
    }
    if(assemble.config.lst_file)
    {
        if (assemble.config.verbose >= 2)
            cout << "write listing to file: " << assemble.config.lst_file << endl;
    }

    if(flog)
    {
        if (assemble.config.verbose >= 2)
            cout << "write error output to file: " << flog << endl;
        errorFile.open(flog);
        if(!errorFile)
            do_error(ERR_FATAL|RETC_FCREATE, 0);
    }

    if(!fout)
    {
        cout << "No output file selected: virtual mode." << endl;
        if(errorFile)
        {
            errorFile << "no output file selected: virtual mode." <<
                endl << endl;
        }
    }

    if(!errorFile)
    {
        if (assemble.config.verbose >= 2)
            cout << endl;
    }

    //start assembling NOW!
    assemble.now(fin, fout);
    return(0);
}
