//=============================================================================
// PARAM CLASS
//=============================================================================

#include "params.h"

#include "labels.h"


//-----------------------------------------------------------------------------
// PARAM::add: adds a param 2 the list
//-----------------------------------------------------------------------------
DWORD PARAM::add(char *param)
{
	dbgout("PARAM::add(CHAR[" << param << "])\n")

	IM_RET imr;		//used DURING call to inetrnal math routine
	PARAMLIST *pl;
	char *chr;

	if(!param) return(RETC_PARAM_EXPECTED);

	//skip spaces or tabs
	while((*param==' ')||(*param=='\t')) param++;
	if(!*param) return(RETC_PARAM_EXPECTED);

	char *last = param+strlen(param)-1;
	while((last>=param)&&((*last==' ')||(*last=='\t'))) last--;
	*(last+1) = NULL;
	if(!*param) return(RETC_PARAM_EXPECTED);

	//add new entry
	if(ParamList->members >= MAX_INSTR_PARAMS) return(RETC_TOO_MUCH_PARAMS);
	pl = (PARAMLIST *)ParamList->add();
	if(!pl) do_error(ERR_FATAL|RETC_LISTALLOC, 0);

	//on PARAM_GET_STRINGS...
	if(flags&PARAM_GET_STRINGS)
	{
		pl->str = (char*)malloc( strlen(param)+1 );
		if(!pl->str) do_error(ERR_FATAL|RETC_OUTOFMEM, 0);
		strcpy(pl->str, param);
	}

	//if virtual label
	if(*param == '\\' && *(param+1) == '@' )
	{
		pl->type = PTYPE_CCNTR;
		return(RETC_OK);
	}

	//if not PARAM_GET_VALUES -> finish
	if(!(flags&PARAM_GET_VALUES)) return(RETC_OK);

	//if param is a string
	if((*param=='"')&&(*last=='"'))
	{
		pl->type = PTYPE_STRING;

		if(!pl->str)	//if not already allocated
		{
			pl->str = (char*)malloc( strlen(param)+1 );
			if(!pl->str) do_error(ERR_FATAL|RETC_OUTOFMEM, 0);
			strcpy(pl->str, param);
		}

		//convert str-data (terminates always)
		param = pl->str;
		chr = param+1;
		while(*chr) switch(*chr)
		{
			case '"': if(*(++chr) == '"') *(param++) = *(chr++);
					  else
					  {
						*param = NULL;
						if(*chr) return(RETC_MARK_MISSING);
						else return(RETC_OK);
					  }
					  break;

			default: *(param++) = *(chr++);
		}
		*param = NULL;
		return(RETC_MARK_MISSING);
	}

	//if param is value
	if(intmath.extract(param, &pl->val) == RETC_OK)
	{
		pl->type = PTYPE_VALUE;
		return(RETC_OK);
	}

	//get registers (or labels)
	LABELLIST *ll = label.get(param);
	if(ll)	//label found
	{
		pl->val = ll->value;
		switch(ll->type)
		{
		case LBL_TYPE_DEFINE:
		case LBL_TYPE_EQUATE:
			pl->type = PTYPE_VALUE;
			break;

		case LBL_TYPE_REG:
			pl->type = PTYPE_REG;
			break;

		case LBL_TYPE_PREDEFINE:
			pl->retc = RETC_LABEL_UNDEF;

		default:
			pl->type = PTYPE_LBL;
		}
	}
	else
	{
		if((*param == '(')&&(*last == ')')) //is it a (base)???
		{
			*last = NULL;
			ll = label.get(param+1);
			if((ll)&&(ll->type == LBL_TYPE_REG))	//yes, it is!!!
			{
				pl->type = PTYPE_DEFOFF;	//set param 2 default off
				//and add the base
				pl = (PARAMLIST *)ParamList->add();
				if(!pl) do_error(ERR_FATAL|RETC_LISTALLOC, 0);
				pl->type = PTYPE_BASE;
				pl->val = ll->value;
				return(RETC_OK);
			}

			*last = ')';	//otherwise restore
		}

		//otherwise test if param is ofs(base)
		if((*last == ')')&&( chr = strrchr(param, '(') ))
		{
			*last = NULL;
			ll = label.get(chr+1);	//test4: (label)
			if((ll)&&(ll->type == LBL_TYPE_REG))	//its a base!
			{
				*chr = NULL;
				pl->retc = RETC_OK;
				imr.str = param;
				if(!intmath.calc(&imr))  pl->retc = imr.imrc;

				pl->type = PTYPE_OFF;
				pl->val = imr.value;

				//and add the base
				pl = (PARAMLIST *)ParamList->add();
				if(!pl) do_error(ERR_FATAL|RETC_LISTALLOC, 0);
				pl->type = PTYPE_BASE;
				pl->val = ll->value;
				pl->retc = RETC_OK;
				return(RETC_OK);
			}

			*last = ')';	//otherwise restore
		}

		//otherwise it should be a value
		imr.str = param;
		if(intmath.calc(&imr))
		{
			pl->type = PTYPE_VALUE;
			pl->val = imr.value;
		}
		else pl->retc = imr.imrc;
	}

	return(RETC_OK);
}


//-----------------------------------------------------------------------------
// PARAM::scan: scans a str 4 params
//-----------------------------------------------------------------------------
DWORD PARAM::scan(char *args)
{
	dbgout("PARAM::scan(CHAR*[" << (DWORD)args << "])\n")

	char *param = NULL;
	int marks=0;	//set if entering a string
	int coma=true;	//set if coma found
	DWORD ret;

	scan_cntr++;

	//backup argument string, don't hurt anything, could be needed later
	if((!args)||(!*args)) return(RETC_OK);
	char *_args = (char *)malloc(strlen(args)+1);
	if(!_args) do_error(ERR_FATAL|RETC_OUTOFMEM, ERR_SHOW_NOTHING);
	strcpy(_args, args);
	args = _args;

	for(;*args;args++)
	{
		switch(*args)
		{
		case '"':	//filter strings ":)", usage of "" -> " should work
			if(marks == 1) marks = 0;
			else if(!marks) marks = 1;
			if(coma) {coma=false; param=args;}
			break;

		case '\'':	//filter strings ':)'
			if(marks == 2) marks = 0;
			else if(!marks) marks = 2;
			if(coma) {coma=false; param=args;}
			break;

		default:	//otherwise scan all that stuff
			if(!marks) switch(*args)
			{
			case ',':
				if(coma) {free(_args); return(RETC_PARAM_EXPECTED);}

				coma = true;	//new param!
				*args = NULL;
				if((ret = add(param)) != RETC_OK) {free(_args); return(ret);}
				param = args;
				break;

			case '\r':case '\n':case ';':
				goto lbl_leave;		//this is the end my friend

			default:
				if((coma)&&(*args!=' ')&&(*args!='\t')) {coma=false; param=args;}
			}
		}
	}

lbl_leave:
	if(marks) ret = RETC_MARK_MISSING;
	else if(coma)
	{
		if(!param) ret = RETC_OK;
		else ret = RETC_PARAM_EXPECTED;
	}
	else
	{
		//read last param
		*args = NULL;
		ret = add(param);
	}

	free(_args);
	return(ret);
}


//-----------------------------------------------------------------------------
// PARAM::constructor
//-----------------------------------------------------------------------------
PARAM::PARAM(WORD set_flags)
{
	dbgout("PARAM::PARAM(WORD[" << set_flags << "])\n")

	scan_cntr = 0;
	flags = set_flags;
	ParamList = new LIST( sizeof(PARAMLIST) );
}


//-----------------------------------------------------------------------------
// PARAM::destructor
//-----------------------------------------------------------------------------
PARAM::~PARAM()
{
	dbgout("PARAM::~PARAM()\n")

	PARAMLIST *pl;

	dbgout("free " << ParamList->members << " params\n")

	for(DWORD cntr = ParamList->members; cntr; cntr--)
	{
		pl = (PARAMLIST *)ParamList->get(cntr-1);
		if(!pl) do_error(ERR_FATAL|RETC_LISTREQUEST, 0);

		if(pl->str)
		{
			free(pl->str);
			pl->str = NULL;
			dbgout("free param string #" << cntr << "\n")
		}
	}

	delete ParamList;
}
